#  Copyright (c) 2021 Food-X Technologies
#
#  This file is part of foodx_devops_tools.
#
#  You should have received a copy of the MIT License along with
#  foodx_devops_tools. If not, see <https://opensource.org/licenses/MIT>.

"""Declarations of constants for pipeline refereance parsing."""

BRANCH_PREFIX = r"refs/heads/"
TAG_PREFIX = r"refs/tags/"

SEMANTIC_VERSION_GITREF = r"(\d+\.\d+\.\d+)"
